use bitboard::Bitboard;
use bitboard::Square;
use bitboard::*;
use search::Game;

pub type Side = bool;

pub const WHITE: Side = false;
pub const BLACK: Side = true;

pub type PieceType = u8;

pub const PAWN: PieceType = 0;
pub const KNIGHT: PieceType = 1;
pub const BISHOP: PieceType = 2;
pub const ROOK: PieceType = 3;
pub const QUEEN: PieceType = 4;
pub const KING: PieceType = 5;

#[derive(Copy, Clone, Debug)]
pub struct SimpleMove {
    from: Square,
    to: Square,
}

#[derive(Debug)]
pub enum Move {
    Default { mov: SimpleMove, piece_type: PieceType },
    Castling { side: Side, left: bool },
    EnPassant { side: Side, column: u8 },
    Promotion { mov: SimpleMove, promote_to: PieceType },
}

/**
 * \brief Iterator to serialize bitboard bit for bit
 *
 * Extracts each bit out of the bitboard and returns each one as a separate bitboard.
 */
pub struct BitboardIterator { pub board: Bitboard }

impl Iterator for BitboardIterator {
    type Item = Bitboard;

    fn next(&mut self) -> Option<Bitboard> {
        if self.board != 0 {
            let lsb = self.board & (0_u64.wrapping_sub(self.board));
            self.board &= !lsb;
            //Some(lsb.trailing_zeros())
            Some(lsb)
        } else {
            None
        }
    }
}


pub fn generate_moves(game: &Game, side: Side) -> Vec<Move> {
    let mut moves: Vec<Move> = Vec::new();
    generate_pawn_pushes(game, side, &mut moves);
    generate_pawn_captures(game, side, &mut moves);
    generate_bishop_moves(game, side, &mut moves);
    return moves;
}


/*
pub fn generate_pawn_moves(game: &Game, side: Side) -> Bitboard {
    let pushed = generate_pawn_pushes(game, side);
    let iter = BitboardIterator { board: pushed };
    for p in iter {
        println!("{}", p.trailing_zeros());
    }

    return 0;
}*/

fn generate_pawn_pushes(game: &Game, side: Side, move_list: &mut Vec<Move>) {
    let pawns = game.pawns(side);
    let others = game.get_all_side(!side);
    let moved_pawns =
        match side {
            WHITE => north_one(pawns),
            BLACK => south_one(pawns)
        } & !(ROW_1 | ROW_8);

    let iter = BitboardIterator { board: moved_pawns & !others };
    for p in iter {
        let origin = match side {
            WHITE => south_one(p),
            BLACK => north_one(p)
        };
        move_list.push(Move::Default { mov: SimpleMove { from: square(origin), to: square(p) }, piece_type: PAWN });
    }
}

fn generate_pawn_captures(game: &Game, side: Side, move_list: &mut Vec<Move>) {
    let pawns = game.pawns(side);
    let others = game.get_all_side(!side);
    let promotion_mask = ROW_1 | ROW_8;

    let pawn_iterator = BitboardIterator { board: pawns };

    for pawn in pawn_iterator {
        let left_cap = match side {
            WHITE => northeast_one(pawn),
            BLACK => southeast_one(pawn)
        };
        let right_cap = match side {
            WHITE => northwest_one(pawn),
            BLACK => southwest_one(pawn)
        };

        for cap in [left_cap, right_cap].iter() {
            if cap & others != 0 {
                let simple_move = SimpleMove { from: pawn.trailing_zeros() as u8, to: cap.trailing_zeros() as u8 };
                if cap & promotion_mask != 0 {
                    move_list.push(Move::Promotion { mov: simple_move, promote_to: QUEEN });
                    move_list.push(Move::Promotion { mov: simple_move, promote_to: ROOK });
                    move_list.push(Move::Promotion { mov: simple_move, promote_to: BISHOP });
                    move_list.push(Move::Promotion { mov: simple_move, promote_to: KNIGHT });
                } else {
                    move_list.push(Move::Default { mov: simple_move, piece_type: PAWN });
                }
            }
        }
    }
}


#[cfg(test)]
mod tests {
    use search::Game;
    use movegen::*;

    #[test]
    fn pawn_pushes() {
        let mut game: Game = Game::default();
        let pawn_moves = generate_pawn_moves(&game, WHITE);
        assert_eq!(pawn_moves, ());
    }
}




